;; standard:  a standard libary for Common Lisp
#|
-------------------------------------------------------------------------
This software is Copyright (c) 2008 Daniel S. Bensen.
Permission is granted to use, copy, modify, and distribute this software,
provided that this copyright and permission notice is included in full
in all copies and supporting documentation.
This software is provided "as is" with no express or implied warranty.
-------------------------------------------------------------------------
|#

;;(defsetf if (test if-true if-false) (expr)
;;  (with-gensym val
;;    `(let ((,val ,expr))
;;      (if ,test (setf ,if-true ,val) (setf ,if-false ,val)))

(cl:declaim (optimize debug))

(cl:in-package :standard)

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defmacro eval-always (&body body)
    `(eval-when (:compile-toplevel :load-toplevel :execute)
      ,@body))

  (defmacro xport (&rest symbols)
    `(eval-when (:load-toplevel :execute) (export ',symbols)))
  (xport eval-always xport)

  (defmacro macx (macro args &body body)
    `(progn
      (defmacro ,macro ,args ,@body)
      (xport ,macro)))
  (xport macx)

  (macx defmac (macro args &body body) `(defmacro ,macro ,args ,@body))

(macx in-defpackage (name &rest options)
  `(progn
     (defpackage ,name ,@options)
     (in-package ,name)))

;;; GENSYMS

(macx with-gensyms (items &body body)
  `(let ,(mapcar (lambda (x) (if (atom x)
				 `(,x (gensym))
				 `(,(car x) (gensym ,(cadr x)))))
		 items)
    ,@body))

(macx with-gensym (item &body body) `(with-gensyms (,item) ,@body))

;;; looping

(macx    while (test &body body) `(loop (unless ,test (return)) ,@body))
(macx    until (test &body body) `(loop ( when  ,test (return)) ,@body))
(macx do-while (test &body body) `(loop ,@body (unless ,test (return))))
(macx do-until (test &body body) `(loop ,@body ( when  ,test (return))))

;;; BINDING

(macx lett    ( x   expr &body body) `( destructuring-bind ,x    ,expr ,@body))
(macx letvals (vars expr &body body) `(multiple-value-bind ,vars ,expr ,@body))

;;; DEFs

(macx def (func args &body defn)
  (let (preamble (body defn))
    (when (listp body)
      (when (stringp (car body)) (push (pop body) preamble))
      (while (and (listp (car body))
		  (eq (caar body) 'declare)) (push (pop body) preamble)))
    (with-gensym (val "GRETURN-VAL")
      (setf body
	    (list `(lett gfuncname (symbol-name ',func)
		    (declare (ignorable gfuncname))
		    (macrolet ((greturn (&optional (,val (values)))
				 `(return-from ,',func ,,val))) ;; C-style return
		      ,@body)))))
    (when preamble
      (setf body (revappend preamble body)))
    `(defun ,func ,args ,@body)))
(xport greturn gfuncname)

(macx   defx (f xs &body body) `(progn                       (def ,f ,xs ,@body) (export ',f)))
(macx indef  (f xs &body body) `(progn (declaim (inline ,f)) (def ,f ,xs ,@body)))
(macx indefx (f xs &body body) `(progn (declaim (inline ,f)) (def ,f ,xs ,@body) (export ',f)))
);end eval-when

(defx gensyms (n) (loop repeat n collect (gensym)))

(macx eif ;Elisp-style IF (if* conflicts w/allegro)
 (test then &body elses) `(if ,test ,then (progn ,@elses)))

;;; bindings

(eval-always
(defx 2list<-plist (plist)
  (loop for var in plist
	and expr in (cdr plist)
	collect (list var expr)))
)
(macx letts (plist &body body)
  (eif (no plist)
      `(progn ,@body)
    (lett bindings (nreverse (2list<-plist plist))
      (assert bindings (bindings) "LETTS: malformed binding list")
      (lett binding (car bindings)
	(setf body `(lett ,(car binding) ,(cadr binding) ,@body))
	(dolist (binding (cdr bindings) body)
	  (setf body `(lett ,(car binding) ,(cadr binding) ,body)))))))

(defx external-symbols (&optional (package *package*))
  (lett symbols nil
    (do-external-symbols (symbol package (nreverse symbols))
      (push symbol symbols))))

;;; abbrevs

(defx no (x) (declare (inline null))  (null x))
(macx concat   (&rest args) `(concatenate  ,@args))
(defx concstr (&rest strings) (apply #'concatenate 'string strings))
(defx ask-yn     (str) (declare (inline  y-or-n-p  )) ( y-or-n-p   str))
(defx ask-yes-no (str) (declare (inline yes-or-no-p)) (yes-or-no-p str))
(macx defparam (&rest args) `(defparameter ,@args))

(macx fn (args &body body) `(lambda  ,args ,@body))
(macx fc (func &rest args) `(funcall ,func ,@args))

(defx debug-out (mesg) mesg);; (write-line mesg))

(defx  echo  (&rest args) (apply #'format t args) (terpri))
(defx prompt (&rest args) (apply #'format t args) (finish-output))

(defx sethash (key table val) (setf (gethash key table) val))

(macx for ((var init test incr) &body body)
      `(lett ,var ,init
	(while ,test
	  ,@body
	  (setf ,var ,incr))))

(macx docells ((cell list &optional rzult) &body body)
      `(do ((,cell ,list (cdr ,cell)))
	(,cell ,rzult)
      ,@body))

(macx dopairs ((x y list) &body body)
      (with-gensym cell
	`(lett ,x (car ,list)
	  (docells (,cell (cdr ,list))
	   (dolist (,y (cdr ,cell))
	     ,@body
	     (setf ,x (car ,cell)))))))

(macx do-hashtable ((key val table &optional (rtrnval nil)) &body body)
  (with-gensyms (get-keyval got-one do-hashtable)
    `(block ,do-hashtable
      (with-hash-table-iterator
	  (,get-keyval ,table) 
	(loop
	 (multival-bind (,got-one ,key ,val) (,get-keyval)
			(when (not ,got-one) (return-from ,do-hashtable ,rtrnval))
			,@body))))))

(defx pause (&rest args)
  (when args (apply #'prompt args))
  (read-line))

(macx dotimes1 ((n nmax &optional result) &body body)
  `(do ((,n 1 (1+ ,n)))
       ((> ,n ,nmax) ,result)
       ,@body))

(macx dorange ((n min top &optional result) &body body)
  `(do ((,n ,min (1+ ,n)))
       ((>= ,n ,top) ,result)
	,@body))

(defx range (n-min n-upr) "[n-min...n-upr]"
  (let ((vals '()))
    (do ((n (1- n-upr) (1- n)))
	((< n n-min) vals)
      (push n vals))))

(defx range* (min max) (range min (1+ max)))

(defx times  (n) "[0...n]" (range  0 n))
(defx times1 (n) "[1..n]"  (range* 1 n))

(macx count-list ((n val list &optional result) &body body)
  (with-gensym cell
    `(do* ((,n 0 (+ ,n 1))
	   (,cell      ,list  (cdr ,cell))
	   (,val  (car ,cell) (car ,cell)))
	  ((null ,cell) ,result)
	  ,@body)))

(defx split-list (pred items)
  (let (goods bads)
    (dolist (item items)
      (if (funcall pred item)
	  (push item goods)
	  (push item bads)))
    (values (nreverse goods)
	    (nreverse bads))))

(defx has-length (list n)
  (assert (listp list)  (list) "has-length: 1st arg must be a list")
  (assert (integerp n)  (n)    "has-length: 2nd arg must be an integer")
  (assert (not (minusp n)) (n) "has-length: length must be non-negative")
  (if (zerop n)
      (not list)
      (lett last-cdr (nthcdr (1- n) list)
	(and last-cdr (not (cdr last-cdr))))))

;; predicates

(defx is-int (x &key val sign)
  (and (integerp x)
       (if val (= x val)
	   (or (not sign)
	       (case sign
		 (:>0 ( plusp x)) (:<=0 (not ( plusp x)))
		 (:<0 (minusp x)) (:>=0 (not (minusp x)))
		 (:=0 ( zerop x)) ((:/=0 :!=0 :<>0) (not ( zerop x))))))))

(defx islist (x &key length min-length max-length)
  (and (listp x)
       (if length (has-length x length)
	   (and (or (not min-length)
		    (and (zerop min-length) (not x))
		    (nthcdr (1- min-length) x))
		(or (not max-length)
		    (not (nthcdr max-length x)))))))

(defx isarray (x &key dims rank)
  (and (arrayp x)
       (if dims
	   (equal dims (array-dimensions x))
	   (or (not rank)
	       (= (array-rank x) rank)))))

(defx xorf (a b) (if a (not b) b))

(defx  orf (&rest vals) (dolist (val vals nil) (when   val (return val))))
(defx andf (&rest vals)
  (lett rzult nil
    (dolist (val vals rzult) (if val (setf rzult val) (return nil)))))

(defx /_ (x y) (floor x y))
(defx /. (x y) (float (/ x y)))

(defmac def-length-op (func op)
  (with-gensyms (a b)
    `(defx ,func (,a ,b) (,op (length ,a) (length ,b)))))

(def-length-op length< <)
(def-length-op length= =)
(def-length-op length> >)

(defx sum  (vals) (reduce #'+ vals))
(defx prod (vals) (reduce #'* vals))

(defx e^  (x) (exp     x))
(defx 2^  (x) (expt  2 x))
(defx 10^ (x) (expt 10 x))

(defx variance (x1 x2 &rest rest)
  (let* ((xs (list* x1 x2 rest))
	 (N (length xs))
	 (avg (/. (sum xs) N))
	 (sum (sum (mapcar (fn (x) (expt (- x avg) 2))
			       xs))))
    (values (/. sum (1- N)) avg)))

(defx rms (x1 x2 &rest rest)
  (sqrt (apply #'variance (list* x1 x2 rest))))

(defmacro def-defthings (defthings defthing)
  `(macx ,defthings (&rest things)
    (let ((defs (mapcar (fn (thing) (if (atom thing)
					`(,',defthing  ,thing)
					`(,',defthing ,@thing)))
			things)))
      `(progn ,@defs))))

(def-defthings defvars      defvar     )
(def-defthings defparams    defparam   )
(def-defthings defconstants defconstant)

(defmacro def-setf-defthings (defthings defthing)
  `(macx ,defthings (&rest things)
    (let ((defs (mapcar (fn (thing) (if (atom thing)
					`(,',defthing  ,thing)
					`(,',defthing ,(car thing))))
			things))
	  (sets (delete nil (mapcar (fn (thing)
				      (and (listp thing)
				       `(setf ,(car thing) ,(cadr thing))))
			things))))
      `(progn ,@defs ,@sets))))

(def-setf-defthings setf-defvars      defvar     )
(def-setf-defthings setf-defparams    defparam   )
(def-setf-defthings setf-defconstants defconstant)

(let ((lcs #(#\a #\b #\c #\d #\e #\f #\g #\h #\i #\j #\k #\l #\m
	     #\n #\o #\p #\q #\r #\s #\t #\u #\v #\w #\x #\y #\z))
      (ucs #(#\A #\B #\C #\D #\E #\F #\G #\H #\I #\J #\K #\L #\M
	     #\N #\O #\P #\Q #\R #\S #\T #\U #\V #\W #\X #\Y #\Z)))
  (defun is-case (letr letrs)
    (dotimes (n 26 nil)
      (when (char= letr (aref letrs n))
	(return t))))
  (defx is-lc (letr) (is-case letr lcs))
  (defx is-uc (letr) (is-case letr ucs))
  (defx is-letr (char) (or (is-lc char) (is-uc char)))
  (defx lc<-int (n) (aref lcs n))
  (defx uc<-int (n) (aref ucs n))
  )

(defx int<-letr (c)
  (case c
	((#\A #\a)  0) ((#\B #\b)  1) ((#\C #\c)  2) ((#\D #\d)  3) ((#\E #\e)  4)
	((#\F #\f)  5) ((#\G #\g)  6) ((#\H #\h)  7) ((#\I #\i)  8) ((#\J #\j)  9)
	((#\K #\k) 10) ((#\L #\l) 11) ((#\M #\m) 12) ((#\N #\n) 13) ((#\O #\o) 14)
	((#\P #\p) 15) ((#\Q #\q) 16) ((#\R #\r) 17) ((#\S #\s) 18) ((#\T #\t) 19)
	((#\U #\u) 20) ((#\V #\v) 21) ((#\W #\w) 22) ((#\X #\x) 23) ((#\Y #\y) 24)
	((#\Z #\z) 25)))

(defun make-? (forms)
  (and forms `(if ,(car forms) ,(cadr forms) ,(make-? (cddr forms)))))

(macx ? (forms) (make-? forms))

(defx flatten (list) (apply #'nconc list))

(macx npushlist   (list place) `(setf ,place (   nconc  ,list ,place)))
(macx pushlist    (list place) `(setf ,place (   append ,list ,place)))
(macx pushrevlist (list place) `(setf ,place (revappend ,list ,place)))

(defx trim-whitespace (string) (string-trim '(#\Space #\Tab #\Newline) string))
(defx trim-linespace  (string) (string-trim '(#\Space #\Tab)           string))

(macx pushwhen (expr place)
  (with-gensym val
    `(lett ,val ,expr
      (when ,val (push ,val ,place)))))

(macx letwhen (var pred &body body)
  `(lett ,var ,pred
     (when ,var ,@body)))

(macx setwhen (oldval place newval)
  `(when (eq ,place ,oldval) (setf ,place ,newval)))

(macx cond-eql (val &rest tests)
  (flet ((make-case (test)
	   (let ((case (car test))
		 (prog (cdr test))) `((eql ,val ,case) ,@prog))))
    (let ((cases
	   (mapcar #'make-case tests)))
      `(cond
	,@cases))))

(macx strcase (expr &rest clauses)
  (with-gensym val
  `(lett ,val ,expr
    (cond
      ,@(mapcar (fn (clause) (if (eq (car clause) t)        `(t ,@(cdr clause))
				 `((string= ,val ,(car clause)) ,@(cdr clause))))
		clauses)))))

(macx dovector ((n val vectr &optional rzult) &body body)
  `(dotimes
    (,n (length ,vectr) ,rzult)
    (let ((,val (aref ,vectr ,n))) ,@body)))

(defx string-or-symbol-name (x)
  (cond ((stringp x) x)
	((symbolp x) (symbol-name x))
	(t (error "string-or-symbol-name: Not a string or symbol"))))

(defun letstruct-testable (testing type- slots expr body)
  "Deconstruct a struct.  Testing for struct existance is optional."
  (lett prefix (string-or-symbol-name type-)
    (with-gensym val
      (flet ((slotbinding (slot)
	       (let* ((is-atom (atom slot))
		      (struct-slot 
		       (find-symbol (concstr prefix
					     (symbol-name (if is-atom slot
							      (cadr slot))))))
		      (slotname	(if is-atom slot (car slot)))
		      (slotexpr1 (list struct-slot val))
		      (slot-expr (if testing
				     `(and ,expr ,slotexpr1)
				     slotexpr1)))
		 (list slotname slot-expr))))
	(if slots
	    `(lett ,val ,expr
	      (let ,(mapcar #'slotbinding slots) ,@body))
	    `(lett ,val ,expr                    ,@body))))))

(macx letstruct    ((type- . slots) x &body body) (letstruct-testable nil type- slots x body))
(macx letstruct-if ((type- . slots) x &body body) (letstruct-testable  t  type- slots x body))

(defx |read_case| (case)
  (lett readtable (copy-readtable nil)
    (setf (readtable-case readtable) case)
    (setf *readtable* readtable)))

(defx cwd () (truename "."))

(defx cd (pathstr)
;;  (lett char0 (char dir 0)
    (setf *default-pathname-defaults*
	  (pathname pathstr)))
	;;  (if [char0 char= #\/]
	  ;;    dir
	    ;;  (concatenate 'string
		;;	   *default-pathname-defaults*
			;;   dir)))))

#|

(defx quit ()
  #+allegro (excl:exit)
  #+sbcl  (sb-ext:quit)
  #+clisp    (ext:quit)
  #-(or allegro sbcl clisp) (error "Don't know how to quit."))

;; regexes
(defmacro match-values-bind ((the-match &rest reg-names) regex target &body body)
  (let ((reg-array (gensym))
	(num-names (gensym))
	(reg-list  (gensym)))
    `(let*
      ((,num-names (length (quote ,reg-names)))
       (,reg-list '()))
      (assert (<= ,num-names 10) () "Too many names:  ~D" ,num-names)
      (multiple-value-bind
	    (,the-match ,reg-array) (scan-to-strings ,regex ,target)
	;;(format t "the-match = ~A~%" ,the-match)
	;;(format t "reg-array = ~A~%" ,reg-array)
	(when ,reg-array
	  (dotimes (n ,num-names) (push (aref ,reg-array n) ,reg-list)))
	(multiple-value-bind
	      ,reg-names (values-list (nreverse ,reg-list))
	  ,@body)
	))))

(let ((regx "([a-z]+)\\W*(\\d+)")
      (target "!@#$#%abc.123^&*()"))
  (match-values-bind
   (the-match letrs nums) regx target
   (assert the-match nil "No match~%")
   (format t "match-values-bind ~S ~S => ~A ~A ~A~%" regx target the-match letrs nums))
  (match-values-bind
   (the-match letrs nums) regx ""
   (assert (not the-match) nil "Shouldn't be a match~%")
   (format t "match-values-bind ~S ~S => ~A ~A ~A~%" regx "" the-match letrs nums))
  )



|#
